<?php
/*
v1.1
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    LJProxy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
require_once('config.php');
date_default_timezone_set('GMT');

class LiveJournalRpc
{
	private $ch = false; /** cURL handle */
	private $friendGroups = array(); /** friend group number => friend group name */
	private $fiendMemberships = array(); /** friend group name => array of friend names */
	private $loggedIn = false; /** have we successfully logged in? */
	private $username = ''; /** LJ username */
	private $password = ''; /** LJ password */
	private $sessionCookie = ''; /** LJ session cookie */
	private $outputThrottle = 100; /** how many entries to put in the output RSS */
	
	public function destroy()
	{
		unlink('cookiefile.txt');
		curl_close($this->ch);
	}
	
	/** Make an LJ RPC request with the given name/value pairs and return the resulting name/value pairs */
	private function makeRequest($pairs)
	{
		$postData = '';
		$result = array();
		$tempFile = tempnam('/tmp', 'ljag');
		$f = fopen($tempFile, 'w');
		reset($pairs);
		while (list($name, $value) = each($pairs))
			$postData .= urlencode($name) . '=' . urlencode($value) . '&';
		//var_dump($postData); print("<br /><br />\n\n");
		if ($this->ch === false)
			$this->ch = curl_init('http://www.livejournal.com/interface/flat');
		curl_setopt($this->ch, CURLOPT_URL, 'http://www.livejournal.com/interface/flat');
		curl_setopt($this->ch, CURLOPT_POST, 1);
		curl_setopt($this->ch, CURLOPT_POSTFIELDS, $postData);
		curl_setopt($this->ch, CURLOPT_USERAGENT, 'LJ Friend Group Aggregator <brian@netninja.com>');
		curl_setopt($this->ch, CURLOPT_FILE, $f);
		/*
		if (strlen($this->sessionCookie) > 0)
		{
			curl_setopt($this->ch, CURLOPT_HTTPHEADER, array('X-LJ-Auth' => 'cookie'));
			curl_setopt($this->ch, CURLOPT_COOKIE, 'ljsession=' . $this->sessionCookie);
		}
		*/
		/*
		$cf = fopen('cookiefile.txt', 'a'); fclose($cf);
		curl_setopt($this->ch, CURLOPT_COOKIESESSION, 0);
        curl_setopt($this->ch, CURLOPT_COOKIEFILE, 'cookiefile.txt');
        curl_setopt($this->ch, CURLOPT_COOKIEJAR, 'cookiefile.txt');
		*/
 		$rc = curl_exec($this->ch);
		fclose($f);
		if (!$rc)
			return false;
		if (($f = fopen($tempFile, 'r')) === false)
			return false;
		while (($name = fgets($f)) !== false)
		{
			$value = fgets($f);
			$result[trim($name)] = trim($value);
		}
		fclose($f);
		unlink($tempFile);
		return $result;
	}
	
	/** 
	 * Get the (authenticated, so includes protected entries) RSS for the given person.
	 * Returns the raw file content as string or false.
	 */
	private function getRss($username)
	{
		$result = '';
		$postData = '';
		$tempFile = tempnam('/tmp', 'ljag');
		$f = fopen($tempFile, 'w');
		$url = 'http://' . $username . '.livejournal.com/data/rss?auth=digest';
		if ($this->ch === false)
			$this->ch = curl_init($url);
		curl_setopt($this->ch, CURLOPT_URL, $url);
		curl_setopt($this->ch, CURLOPT_USERAGENT, 'LJ Friend Group Aggregator <brian@netninja.com>');
		curl_setopt($this->ch, CURLOPT_FILE, $f);
		curl_setopt($this->ch, CURLOPT_FOLLOWLOCATION, 1); 
		curl_setopt($this->ch, CURLOPT_HTTPAUTH, CURLAUTH_DIGEST); 
		curl_setopt($this->ch, CURLOPT_UNRESTRICTED_AUTH, 1);
		curl_setopt($this->ch, CURLOPT_USERPWD, $this->username . ':' . $this->password);
 		$rc = curl_exec($this->ch);
		fclose($f);
		if (!$rc)
			return false;
		if (($f = fopen($tempFile, 'r')) === false)
		fclose($f);
		$result = file_get_contents($tempFile);
		unlink($tempFile);
		//$this->log($result);
		return $result;
	}
	
	/** Login with the given username/password and retrieve friends and friend groups. Returns true/false for success. */
	private function loginInternal($username, $password)
	{
		$params = array('mode' => 'getchallenge');
		$result = $this->makeRequest($params);
		//var_dump($result); print("<br /><br />\n\n");
		$this->username = $username;
		$this->password = $password;
		$challenge = $result['challenge'];
		$passwordHash = md5($challenge . md5($this->password, false), false);
		//var_dump($challenge); print("<br /><br />\n\n");
		$params = array('mode' => 'login');
		$params['user'] = $this->username;
		$params['auth_method'] = 'challenge';
		$params['auth_challenge'] = $challenge;
		$params['auth_response'] = $passwordHash;
		$params['clientversion'] = 'lj_friend_group_rss/1.0';
		$result = $this->makeRequest($params);
		//var_dump($result); print("<br /><br />\n\n");
		if (strcmp($result['success'], 'OK') == 0)
		{
			$friendGroupCount = floor($result['frgrp_maxnum']);
			$this->friendGroups = array();
			for ($i = 1; $i <= $friendGroupCount; $i++)
				if (array_key_exists('frgrp_' . $i . '_name', $result))
					if (strlen($result['frgrp_' . $i . '_name']) > 0)
					{
						$groupName = $result['frgrp_' . $i . '_name'];
						$this->friendGroups[$i] = $groupName;
						$this->friendMemberships[$groupName] = array();
					}
			//var_dump($this->friendGroups); print("<br /><br />\n\n");
			$this->loggedIn = true;
			return true;
		}
		return false;
	}
	
	/** Load friends and friend groups into the member variables */
	private function getFriends()
	{
		$params = array('mode' => 'getchallenge');
		$result = $this->makeRequest($params);
		//var_dump($result); print("<br /><br />\n\n");
		$challenge = $result['challenge'];
		$passwordHash = md5($challenge . md5($this->password, false), false);
		//var_dump($challenge); print("<br /><br />\n\n");
		$params = array('mode' => 'getfriends');
		$params['user'] = $this->username;
		$params['auth_method'] = 'challenge';
		$params['auth_challenge'] = $challenge;
		$params['auth_response'] = $passwordHash;
		$result = $this->makeRequest($params);
		//var_dump($result); print("<br /><br />\n\n");
		if (strcmp($result['success'], 'OK') != 0)
			return false;
		$friendCount = $result['friend_count'];
		for ($i = 1; $i <= $friendCount; $i++)
		{
			$friendName = $result['friend_' . $i . '_user'];
			// Skip suspended, deleted, etc.
			if (array_key_exists('friend_' . $i . '_status', $result))
				continue;
			// Skip syndicated; LJ just redirects us to the original source's RSS, which may by
			// some other format (different RSS version, Atom, etc. and will likely kill our
			// parser)
			if (array_key_exists('friend_' . $i . '_type', $result))
				if (strcmp($result['friend_' . $i . '_type'], "syndicated") == 0)
					continue;
			// Skip OpenID
			if (array_key_exists('friend_' . $i . '_identity_type', $result))
				continue;
			// Skip not-in-group
			if (!array_key_exists('friend_' . $i . '_groupmask', $result))
				continue;
			$groupMask = floor($result['friend_' . $i . '_groupmask']);
			if (($groupMask != 0) && ($groupMask != 1))
				$this->addFriendToGroups($friendName, $groupMask);
		}
		return true;
	}
	
	/** Given the friend and group bitmask, add the friend to the appropriate groups in
	    this object's member variables */
	private function addFriendToGroups($username, $groupMask)
	{
		//print("<br />\n" . $username . ":<br />\n");
		for ($i = 1; $i < 32; $i++)
		{
			if ((($groupMask >> $i) & 1) == 0)
				continue;
			$groupName = $this->friendGroups[$i];
			array_push($this->friendMemberships[$groupName], $username);
			//print(" : " . $groupName . "<br />\n");
		}
	}

	/** Login and get friends+groups */
	public function login($username, $password)
	{
		if (!$this->loginInternal($username, $password))
			return false;
		if (!$this->makeSession())
			return false;
		return $this->getFriends($username);
	}
	
	/** Return a list of the friend group names */
	public function getFriendGroupNames()
	{
		if ($this->loggedIn)
		{
			$result = array_values($this->friendGroups);
			sort($result);
			return $result;
		}
		return false;
	}
	
	/** For the given group name, return an array of usernames in that group */
	public function getMembersOfGroup($groupName)
	{
		$result = $this->friendMemberships[$groupName];
		if ($result)
			sort($result);
		return $result;
	}
	
	/** Sanitize a group name to something filesystem-friendly */
	public function sanitizeGroupName($groupName)
	{
		$result = '';
		for ($i = 0; $i < strlen($groupName); $i++)
		{
			$c = strtolower(substr($groupName, $i, 1));
			if (strchr("1234567890abcdefghijklmnopqrstuvwxyz", $c) !== false)
				$result .= $c;
			else
				$result .= '_';
		}
		return $result;
	}
	
	/** Sanitize an string to XML entities.  This is really conservative and errs to the
	    side of over-encoding. */
	public function sanitizeXml($s)
	{
		$result = '';
		for ($i = 0; $i < strlen($s); $i++)
		{
			$c = substr($s, $i, 1);
			if (strchr("1234567890abcdefghijklmnopqrstuvwxyz ", strtolower($c)) !== false)
				$result .= $c;
			else
				$result .= '&#' . ord($c) . ';';
		}
		return $result;
	}
	
	/** For a given username, get all recent entries and return them as an associative array
	with timestamp in key and XML of entry in value */
	private function getEntries($username)
	{
		$entries = array();
		$friendRss = $this->getRss($username);
		try {
			$xml = new SimpleXmlElement($friendRss);
			$entryNodes = $xml->xpath('//item');
			$imageNode = $xml->xpath('/rss/channel/image/url');
			if (sizeof($imageNode > 0))
			{
				$imageUrl = (string) $imageNode[0];
				$imageUrl = preg_replace('/<.*?>/', '', $imageUrl);
			} else {
				$imageUrl = '';
			}
			//print("<pre>"); var_dump($entryNodes); print("</pre>");
			for ($i = 0; $i < sizeof($entryNodes); $i++)
			{
				$node = $entryNodes[$i];
				$date = date_parse($node->pubDate);
				$security = '';
				$dateString = $node->pubDate;
				$dateString = preg_replace('/<.*?>/', '', $dateString);
				//print("<pre>"); var_dump($node); print("</pre>");
				// Get the security type (public/protected)
				$securityNode = $node->xpath("lj:security");
				if (sizeof($securityNode) > 0)
					$security = $securityNode[0]->asXml();
				$security = preg_replace('/<.*?>/', '', $security);
				if (strcmp($security, "usemask") == 0)
					$security = 'friends';
				// If not public, make it caps
				if (strcmp($security, "public") != 0)
				{
					$dingbat = '&#x26A0;';
					$security = strtoupper($security);
				} else {
					$dingbat = '';
				}
				$title = (string) $node->title;
				if (strlen($title) == 0)
				{
					$title = (string) $node->pubDate;
					$node->addChild('title', $title);
				}
				// Prepend to title so we'll see it in the RSS
				$title = $this->sanitizeXml($username) . ' - ' . $dingbat . $this->sanitizeXml($security) . $dingbat . " - " . $this->sanitizeXml($title);
				$timestamp = mktime($date['hour'], $date['minute'], $date['second'], $date['month'], $date['day'], $date['year']);
				$entries[$timestamp] = $node->asXml();
				// Replace the <title>...</title> text
				if (strlen($title) > 0)
					$entries[$timestamp] = preg_replace('/<title>.*?<.title>/', '<title>' . $title . '</title>', $entries[$timestamp]);
				$header = '';
				if (strlen($imageUrl) > 0)
					$header .= '<img src="' . $imageUrl . '" align="left" style="margin:0 5px 5px 0;" />';
				$header .= '<small>' . $dateString . '</small><br />';
				$entries[$timestamp] = preg_replace('/<description>/', '<description>' . $this->sanitizeXml($header), $entries[$timestamp]);
			}
		} catch (Exception $e) {
		}
		return $entries;
	}
	
	/** Make a session cookie */
	private function makeSession()
	{
		$params = array('mode' => 'getchallenge');
		$result = $this->makeRequest($params);
		//var_dump($result); print("<br /><br />\n\n");
		$challenge = $result['challenge'];
		$passwordHash = md5($challenge . md5($this->password, false), false);
		//var_dump($challenge); print("<br /><br />\n\n");
		$params = array('mode' => 'sessiongenerate');
		$params['user'] = $this->username;
		$params['auth_method'] = 'challenge';
		$params['auth_challenge'] = $challenge;
		$params['auth_response'] = $passwordHash;
		$params['expiration'] = 'short';
		$result = $this->makeRequest($params);
		//var_dump($result); print("<br /><br />\n\n");
		if (strcmp($result['success'], 'OK') != 0)
			return false;
		$this->sessionCookie = $result['ljsession'];
		return true;
	}
	
	/** For the given group name, load all friends, aggregate entries, and write resulting
	    RSS to file */
	public function makeRss($groupName, $filename)
	{
		$entries = array();
		// Check state
		if (!$this->loggedIn)
			return false;
		// Get a session cookie
		if (!$this->makeSession())
			return false;
		// Open file, write preamble
		$this->clearLog();
		$timestampStart = time();
		$friends = $this->getMembersOfGroup($groupName);
		for ($i = 0; $i < sizeof($friends); $i++)
		{
			$friend = $friends[$i];
			$this->log("Getting " . $friend . " (" . ($i + 1) . " of " . sizeof($friends) . ")");
			$friendEntries = $this->getEntries($friend);
			// Possible (but rare) bug such that if two people published at the exact same second,
			// one will get overwritten by the other.
			$keys = array_keys($friendEntries);
			for ($j = 0; $j < sizeof($keys); $j++)
				$entries[$keys[$j]] = $friendEntries[$keys[$j]];
			$this->log("    Got " . $friend);
		}
		$this->log("Done scraping!");
		$timestampEnd = time();
		$delta = $timestampEnd - $timestampStart;
		$generatedTime = '';
		if ($delta > 60)
			$generatedTime = floor($delta / 60) . ' minute(s), ';
		$generatedTime .= floor($delta % 60) . ' second(s)';
		$f = fopen($filename, "w");
		fwrite($f, "<?xml version='1.0' encoding='utf-8' ?>\n" .
			"<rss version='2.0' xmlns:lj='http://www.livejournal.org/rss/lj/1.0/' xmlns:media='http://search.yahoo.com/mrss/' xmlns:atom10='http://www.w3.org/2005/Atom'>\n" .
			"<channel>\n" .
  			"<title>LJ: " . $groupName . ", " . $this->username . "</title>\n" .
  			"<link>http://livejournal.com/</link>\n" .
  			"<description>LiveJournal RSS: " . $groupName . ", " . $this->username . ", generated in " . $generatedTime . "</description>\n" .
			"<lastBuildDate>" . date('r') . "</lastBuildDate>\n" .
  			"<generator>LiveJournal / LiveJournal.com</generator>\n" .
			"<ttl>60</ttl>\n" .
  			//"<image>\n" .
    		//"<url>http://l-userpic.livejournal.com/27154046/799118</url>\n" .
    		//"<title>now with pureed penguin</title>\n" .
    		//"<link>http://feedle.livejournal.com/</link>\n" .
    		//"<width>100</width>\n" .
    		//"<height>87</height>\n" .
  			//"</image>\n" .
			"\n\n\n"
			);
		$throttle = $this->outputThrottle;
		$keys = array_keys($entries);
		rsort($keys);
		for ($i = 0; $i < sizeof($keys); $i++)
		{
			fwrite($f, $entries[$keys[$i]] . "\n\n");
			if ($throttle-- <= 0)
				break;
		}
		// Write closing tags
		fwrite($f, "</channel></rss>");
		fclose($f);
		$this->log("Done!");
	}
	
	/** Unit test */
	public function testGetEntries($username)
	{
		if (!$this->makeSession())
			return false;
		$entries = $this->getEntries($username);
		print("<pre>"); var_dump($entries); print("</pre>");
	}
	
	/** Clear the debug log file */
	private function clearLog()
	{
		$log = fopen("log.txt", "w"); fclose($log);
	}
	
	/** Add a line to the debug log file */
	private function log($s)
	{
		$log = fopen("log.txt", "a"); 
		fwrite($log, $s . "\n"); 
		fclose($log);
	}
}

/** Unit test */
function testLogin()
{
	GLOBAL $LJ_USERNAME, $LJ_PASSWORD;
	$ljRpc = new LiveJournalRpc();
	$ljRpc->login($LJ_USERNAME, $LJ_PASSWORD);
	$groupNames = $ljRpc->getFriendGroupNames();
	print("<ul>");
	for ($i = 0; $i < sizeof($groupNames); $i++)
	{
		print("<li>" . $groupNames[$i] . "<br />\n");
		$members = $ljRpc->getMembersOfGroup($groupNames[$i]);
		for ($j = 0; $j < sizeof($members); $j++)
			print($members[$j] . " ");
		print("</li>\n");
	}
	print("</ul>\n");
}

?>
